from statsmodels.tsa.seasonal import seasonal_decompose
from statsmodels.graphics.tsaplots import plot_acf, plot_pacf
from statsmodels.tsa.stattools import adfuller

import matplotlib.pyplot as plt
plt.rc('axes', titlesize='large')    
plt.rc('axes', labelsize='large')   
plt.rc('xtick', labelsize='large')   
plt.rc('ytick', labelsize='large')   
plt.rc('legend', fontsize='large')   
plt.rc('figure', titlesize='x-large') 
import seaborn as sns
sns.set_theme(style='white', font_scale=1)


def plot_time_series(series, title = 'Time Series Plot', xlabel = 'Date', ylabel = 'Value'):
    plt.figure(figsize=(10, 6))
    plt.plot(series)
    plt.title(title)
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    plt.show()

def plot_acf_pacf(series, lags=40, title='ACF and PACF Plots'):
    fig, ax = plt.subplots(1, 2, figsize=(16, 6))
    plot_acf(series, ax=ax[0], lags=lags)
    plot_pacf(series, ax=ax[1], lags=lags)
    fig.suptitle(title)
    plt.show()

def test_stationarity(series):
    result = adfuller(series.dropna())
    print('ADF Statistic: %f' % result[0])
    print('p-value: %f' % result[1])
    print('Critical Values:')
    for key, value in result[4].items():
        print('\t%s: %.3f' % (key, value))
    # print whether the series is stationary or not at 5%
    print("Is the series stationary? {0}".format('Yes' if result[1] < 0.05 else 'No'))

def decompose_series(series, period=12, title='Time Series Decomposition'):
    decomposition = seasonal_decompose(series, model='additive', period=period)
    fig = decomposition.plot()
    fig.set_size_inches(10, 8)
    fig.suptitle(title)
    plt.show()
    return decomposition

# Plot yearly seasonality
def plot_yearly_seasonality(series, title='Yearly Seasonality', show_all = True):
    df = series.to_frame("value")
    df['year'] = df.index.year
    df['day_of_year'] = df.index.dayofyear

    yearly_patterns = df.pivot_table(values='value', index='day_of_year', columns='year')

    plt.figure(figsize=(12, 6))
    plt.title(title)
    plt.xlabel("Day of Year")
    plt.ylabel("Values")
    if show_all:
        for column in yearly_patterns:
            plt.plot(yearly_patterns.index, yearly_patterns[column], 
                     label=f'Year {column}', marker='o', linestyle='-')
        plt.legend(title="Year", bbox_to_anchor=(1.05, 1), loc='upper left', frameon=False)
    else:
        plt.plot(yearly_patterns.index, yearly_patterns.mean(axis=1),
                  label='Average', marker='o', linestyle='-')
        plt.legend(frameon=False)
    plt.show()

# Remove yearly seasonality and show monthly patterns
def analyze_monthly_pattern(series, title='Monthly Seasonality', show_all = False):
    df = series.to_frame("value")
    df['month'] = df.index.month
    df['day_of_month'] = df.index.day

    # Adjust for yearly seasonality by subtracting the mean for each day of the year
    daily_means = df.groupby(df.index.dayofyear)['value'].transform('mean')
    df['adjusted'] = df['value'] - daily_means

    monthly_patterns = df.pivot_table(values='adjusted', index='day_of_month', columns='month')

    plt.figure(figsize=(12, 6))
    plt.title(title)
    plt.xlabel("Day of Month")
    plt.ylabel("Adjusted Values")
    if show_all:
        for column in monthly_patterns:
            plt.plot(monthly_patterns.index, monthly_patterns[column], 
                     label=f'Month {column}', marker='o', linestyle='-')
        plt.legend(title="Month", bbox_to_anchor=(1.05, 1), loc='upper left', frameon=False)
    else:
        plt.plot(monthly_patterns.index, monthly_patterns.mean(axis=1),
                 label='Average', marker='o', linestyle='-')
        plt.legend(frameon=False)
    plt.show()

# Further adjust for monthly seasonality and analyze weekly patterns
def analyze_weekly_pattern(series, title='Weekly Seasonality'):
    df = series.to_frame("value")
    df['day_of_week'] = df.index.dayofweek

    # Adjust for yearly seasonality
    daily_means = df.groupby(df.index.month)['value'].transform('mean')
    df['adjusted'] = df['value'] - daily_means

    # Further adjust for monthly seasonality
    monthly_means = df.groupby(df.index.day)['adjusted'].transform('mean')
    df['monthly_adjusted'] = df['adjusted'] - monthly_means

    weekly_patterns = df.pivot_table(values='monthly_adjusted', index='day_of_week')
    day_map = {0: 'Mon', 1: 'Tue', 2: 'Wed', 3: 'Thu', 4: 'Fri', 5: 'Sat', 6: 'Sun'}

    plt.figure(figsize=(10, 5))
    plt.title(title)
    plt.xlabel("Day of Week")
    plt.ylabel("Adjusted Values")
    plt.plot(weekly_patterns.index, weekly_patterns['monthly_adjusted'], marker='o', linestyle='-')
    plt.xticks(weekly_patterns.index, [day_map[day] for day in weekly_patterns.index])
    plt.grid(True)
    plt.show()

def analyse_series(series, series_name ='', period=12, acf_lags=40):
    
    plot_time_series(series, title=f'{series_name}')

    # Stationarity test
    print("Results of Dickey-Fuller Test:")
    test_stationarity(series)
    
    # ACF and PACF plots
    plot_acf_pacf(series, lags=acf_lags, title=f'{series_name} ACF and PACF')
   
    # Decompose series
    decompose_series(series, period=period, title=f'{series_name} Decomposition')

    # Analyze yearly seasonality
    plot_yearly_seasonality(series, title=f'{series_name} Yearly Seasonality')

    # Analyze monthly pattern
    analyze_monthly_pattern(series, title=f'{series_name} Monthly Pattern')

    # Analyze weekly pattern
    analyze_weekly_pattern(series, title=f'{series_name} Weekly Pattern')




import numpy as np
import pandas as pd

import matplotlib.pyplot as plt
import seaborn as sns
plt.rc('axes', titlesize='large')    
plt.rc('axes', labelsize='large')   
plt.rc('xtick', labelsize='large')   
plt.rc('ytick', labelsize='large')   
plt.rc('legend', fontsize='large')   
plt.rc('figure', titlesize='x-large') 
sns.set_theme(style='white', font_scale=1)

from matplotlib.cbook import flatten
from matplotlib.ticker import FuncFormatter, MaxNLocator

from sktime.utils.validation._dependencies import _check_soft_dependencies
from sktime.utils.validation.forecasting import check_interval_df, check_y
from sktime.utils.validation.series import check_consistent_index_type
from sktime.datatypes import convert_to
from typing import List
from warnings import simplefilter, warn


def _check_colors(colors, n_series):
    """Verify color list is correct length and contains only colors."""
    from matplotlib.colors import is_color_like

    if n_series == len(colors) and all([is_color_like(c) for c in colors]):
        return True
    warn(
        "Color list must be same length as `series` and contain only matplotlib colors"
    )
    return False

def plot_series(
    *series,
    labels=None,
    markers=None,
    colors=None,
    title=None,
    xlabel=None,
    ylabel=None,
    ax=None,
	figsize =(15,6),
    pred_interval=None,
	interval_label ='prediction interval'
):
    """
    Plot one or more time series.

    Parameters:
    -----------    
        series : pd.Series or iterable of pd.Series
            One or more time series
        labels : list, default = None
            Names of series, will be displayed in figure legend
        markers: list, default = None
            Markers of data points, if None the marker "o" is used by default.
            The length of the list has to match with the number of series.
        colors: list, default = None
            The colors to use for plotting each series. Must contain one color per series
        title: str, default = None
            The text to use as the figure's suptitle
        pred_interval: pd.DataFrame, default = None
            Output of `forecaster.predict_interval()`. Contains columns for lower
            and upper boundaries of confidence interval.

    Returns:
    --------    
        fig : plt.Figure
        ax : plt.Axis

    Examples:
    ---------    
    >>> from sktime.utils.plotting import plot_series
    >>> from sktime.datasets import load_airline
    >>> y = load_airline()
    >>> fig, ax = plot_series(y) 
    """
    
    _check_soft_dependencies("matplotlib", "seaborn")
    
    for y in series:
        check_y(y)

    series = list(series)
    series = [convert_to(y, "pd.Series", "Series") for y in series]

    n_series = len(series)
    _ax_kwarg_is_none = True if ax is None else False
    # labels
    if labels is not None:
        if n_series != len(labels):
            raise ValueError(
                """There must be one label for each time series,
                but found inconsistent numbers of series and
                labels."""
            )
        legend = True
    else:
        labels = ["" for _ in range(n_series)]
        legend = False

    # markers
    if markers is not None:
        if n_series != len(markers):
            raise ValueError(
                """There must be one marker for each time series,
                but found inconsistent numbers of series and
                markers."""
            )
    else:
        markers = ["o" for _ in range(n_series)]

    # create combined index
    index = series[0].index
    for y in series[1:]:
        # check index types
        check_consistent_index_type(index, y.index)
        index = index.union(y.index)

    # generate integer x-values
    xs = [np.argwhere(index.isin(y.index)).ravel() for y in series]

    # create figure if no Axe provided for plotting
    if _ax_kwarg_is_none:
        fig, ax = plt.subplots(1, figsize=figsize)

    # colors
    if colors is None or not _check_colors(colors, n_series):
        colors = sns.color_palette("colorblind", n_colors=n_series)

    # plot series
    for x, y, color, label, marker in zip(xs, series, colors, labels, markers):
        # scatter if little data is available or index is not complete
        # if len(x) <= 3 or not np.array_equal(np.arange(x[0], x[-1] + 1), x):
        #     plot_func = sns.scatterplot
        # else:
        plot_func = sns.lineplot

        plot_func(x=x, y=y, ax=ax, marker=marker, label=label, color=color)

    # combine data points for all series
    xs_flat = list(flatten(xs))

    # set x label of data point to the matching index
    def format_fn(tick_val, _):
        if int(tick_val) in xs_flat:
            return index[int(tick_val)]
        else:
            return ""

    # dynamically set x label ticks and spacing from index labels
    ax.xaxis.set_major_formatter(FuncFormatter(format_fn))
    ax.xaxis.set_major_locator(MaxNLocator(integer=True))
	
	# Change the xtickslabels
    reformat_xticks = lambda x: x if x=='' else pd.to_datetime(x).strftime('%Y-%m-%d')
    xticklabels = [reformat_xticks(item.get_text()) for item in ax.get_xticklabels()]
    ax.set_xticklabels(xticklabels)
	
    # Set the figure's title
    if title is not None:
        ax.set_title(title, size="xx-large")

    # Label the x and y axes
    if xlabel is not None:
        ax.set_xlabel(xlabel)

    _ylabel = ylabel if ylabel is not None else series[0].name
    ax.set_ylabel(_ylabel)

    if legend:
        ax.legend(loc='best', frameon=False)
	
    if pred_interval is not None:
        check_interval_df(pred_interval, series[-1].index)
        ax = plot_interval(ax, pred_interval, interval_label)
    
    if _ax_kwarg_is_none:
        return fig, ax
    else:
        return ax

def plot_interval(ax, interval_df, legend_label ='prediction interval'):
    cov = interval_df.columns.levels[1][0]
    var_name = interval_df.columns.levels[0][0]
    ax.fill_between(
        ax.get_lines()[-1].get_xdata(),
        interval_df[var_name][cov]["lower"].astype("float64"),
        interval_df[var_name][cov]["upper"].astype("float64"),
        alpha=0.2,
        color=ax.get_lines()[-1].get_c(),
        label=f"{int(cov * 100)}% {legend_label}",
    )
    ax.legend(frameon=False, loc = 'best')
    return ax

def plot_windows(y: pd.Series,
                train_windows: List[np.ndarray],
                test_windows: List[np.ndarray],
                ax: plt.Axes,
                labels: List[str],
                ylabel: str,
                xlabel: str,
                title: str) -> None:
    """
    Visualize training and test windows.

    Parameters:
    -----------        
        train_windows : List[np.ndarray]
            List of training window indices.
        test_windows : List[np.ndarray]
            List of test window indices.
        ax : plt.Axes
            The Axes object for the plot.
        labels : List[str]
            Labels for the plot.
        ylabel : str
            Label for the y-axis.
        xlabel : str
            Label for the x-axis.
        title : str
            The title of the plot.

    Returns:
    --------
        fig : plt.Figure
            If ax was None, a new figure is created and returned
            If ax was not None, the same ax is returned with plot added
        ax : plt.Axis             
            Axes containing the plot 
    """
    assert len(labels)==2, 'wrong number of labels'
    simplefilter("ignore", category=UserWarning)

    def get_y(length, split):
        # Create a constant vector based on the split for y-axis."""
        return np.ones(length) * split
    # reformat_xticks = lambda x: x if x=='' else pd.to_datetime(x).strftime('%Y-%m-%d')

    n_splits = len(train_windows)
    n_timepoints = len(y)
    len_test = len(test_windows[0])

    train_color, test_color = sns.color_palette("colorblind")[:2]

    if ax is None:
        f, ax = plt.subplots(figsize=plt.figaspect(0.5))

    for i in range(n_splits):
        train = train_windows[i]
        test = test_windows[i]

        ax.plot(
            np.arange(n_timepoints), get_y(n_timepoints, i), marker="o", c="lightgray"
        )
        ax.plot(
            train,
            get_y(len(train), i),
            marker="o",
            c=train_color,
            label=labels[0],
        )
        ax.plot(
            test,
            get_y(len_test, i),
            marker="o",
            c=test_color,
            label=labels[1],
        )
    
    xticklabels = [int(item) for item in ax.get_xticks()[1:-1]]
    xticklabels = [''] + [y.index[i].strftime('%Y-%m-%d') for i in xticklabels] + ['']
    ax.invert_yaxis()
    ax.yaxis.set_major_locator(MaxNLocator(integer=True))
    ax.set(
        ylabel=ylabel,
        xlabel=xlabel,
        xticklabels = xticklabels,
    )
    ax.set_title(title, size="xx-large")
    # remove duplicate labels/handles
    handles, labels = [(leg[:2]) for leg in ax.get_legend_handles_labels()]
    ax.legend(handles, labels, frameon=False)

    if ax is None:			
        return f, ax
    else:
        return ax




""" Timeseries Forecasting with Insample, Validation and Out-of-Sample predictions"""
__description__ = "Time Series Forecast"
__author__ = "Amine Raboun - amineraboun@github.io"

# Filter Warnings
import logging
logging.getLogger().setLevel(logging.ERROR)
import warnings
warnings.simplefilter('ignore')
warnings.filterwarnings("ignore")

##############################################################################
# Import Libraries & default configuration 
##############################################################################
# Standard Inports
from typing import Optional, Tuple, List
import pandas as pd
import numpy as np

# Iteration tools
from collections.abc import Iterable
from multiprocessing import Pool
from tqdm import tqdm
import time

# Plotting packages & configuration
from forecast_combine.utils.plotting import plot_series, plot_windows
import matplotlib.pyplot as plt
plt.rc('axes', titlesize='large')    
plt.rc('axes', labelsize='large')   
plt.rc('xtick', labelsize='large')   
plt.rc('ytick', labelsize='large')   
plt.rc('legend', fontsize='large')   
plt.rc('figure', titlesize='x-large') 
import seaborn as sns
sns.set_theme(style='white', font_scale=1)

# Cross Validation tools
from sktime.forecasting.base import ForecastingHorizon
from sktime.forecasting.model_evaluation import evaluate
from sktime.split import CutoffSplitter
from sktime.forecasting.model_selection import (
ExpandingWindowSplitter, 
temporal_train_test_split
)

# Performance Metrics
from .utils.metrics import summary_perf, calculate_prediction_interval
from sktime.performance_metrics.forecasting import (
MeanSquaredError,
MeanAbsoluteError, 
MeanAbsolutePercentageError, 
MedianAbsoluteError
)

##############################################################################
# Default values
##############################################################################
# Common Forcasting models
from sktime.forecasting.naive import NaiveForecaster
from sktime.forecasting.fbprophet import Prophet
from sktime.forecasting.statsforecast import (
StatsForecastAutoARIMA,
StatsForecastAutoETS, 
StatsForecastAutoCES,
StatsForecastAutoTheta,
StatsForecastAutoTBATS, 
)
CommonForecastingModels = {
"Naive": NaiveForecaster(),
"Seasonal_Naive": NaiveForecaster(sp = 5),
"AutoARIMA": StatsForecastAutoARIMA(),
"AutoETS": StatsForecastAutoETS(),
"AutoCES": StatsForecastAutoCES(),
"AutoTheta": StatsForecastAutoTheta(),
"AutoTBATS": StatsForecastAutoTBATS(seasonal_periods = 1),
"Prophet": Prophet(),
}
# Pandas frequencies
pandas_frequency_dict = {
    "D": "daily",
    "B": "business days",
    "W": "weekly (end of week, default on Sunday)",
    "M": "monthly (end of month)",
    "Q": "quarterly (end of quarter)",
    "A": "annual (end of year)",
    "H": "hourly",
    "T": "minutely",
    "S": "secondly",
    "L": "millisecondly",
    "U": "microsecondly",
    "BQ": "business quarterly (business quarter-end)",
    "BA": "business annual (business year-end)",
    "BH": "business hourly (business hour)"
}

##############################################################################
# Master Class Definition and Plot
##############################################################################
class Forecast(object):
    """
    Forecast class for managing and evaluating forecast models.

    Parameters:
    -----------
    - data (pd.DataFrame): A DataFrame containing the input data for forecasting.
    - depvar_str (str): The column name representing the dependent variable for forecasting.
    - fh (int): The forecast horizon, i.e., the number of periods ahead to forecast.
    - pct_initial_window (float): The percentage of data used as the initial training window.
    - step_length (int): The step size for expanding window cross-validation.
    - forecaster_name (str, optional): The name of the forecasting model. Default is 'Naive'.
    - forecaster (object, optional): The forecasting model object. Default is None, which will use the model corresponding to forecaster_name.
    - exog_l (list, optional): List of exogenous variables for forecasting. Default is None.
    - freq (str, optional): The frequency of the time series data. Default is 'B' (business days).

    Attributes:
    -----------
    - y (pd.Series): The time series data representing the dependent variable.
    - X (pd.DataFrame or None): The DataFrame containing exogenous variables or None if there are no exogenous variables.
    - forecaster_name (str): The name of the forecasting model used.
    - forecaster (object): The forecasting model object used for forecasting.
    - fh (ForecastingHorizon): The forecast horizon, i.e., the number of periods ahead to forecast.
    - initial_window (int): The size of the initial training window.
    - step_length (int): The step size for expanding window cross-validation.
    - cv (ExpandingWindowSplitter): The cross-validation window used for expanding window validation.
    - X_train (pd.DataFrame or None): The DataFrame containing exogenous variables for the training set or None if there are no exogenous variables.
    - X_test (pd.DataFrame or None): The DataFrame containing exogenous variables for the test set or None if there are no exogenous variables.
    - y_train (pd.Series): The dependent variable values for the training set.
    - y_test (pd.Series): The dependent variable values for the test set.
    - is_fitted (bool): True if the forecaster is fitted, False otherwise.
    - is_evaluated (bool): True if the forecaster is evaluated on the test set, False otherwise.
    - rs (numpy.random.RandomState): Random state for reproducibility.
    - plot (ForecastPlot): An instance of the ForecastPlot class for plotting utility.
    """

    # Initializer
    def __init__(self,
                 data: pd.DataFrame,
                 depvar_str: str,
                 fh: int,
                 pct_initial_window: float,
                 step_length: int,
                 forecaster_name: Optional[str] = 'Naive',
                 forecaster: Optional[object] = None,
                 exog_l: Optional[list] = None,
                 freq: Optional[str] = 'D',
                 ) -> None:
        """Initializes the Forecast class with the provided parameters."""
        self.__init_test(data, depvar_str, exog_l, freq)
        self.y, self.X = self.__clean_data(data=data, 
                                           depvar_str=depvar_str,
                                           exog_l=exog_l,
                                           freq=freq)
        # Forecasting Model Parameters
        self.forecaster_name = forecaster_name

        if forecaster is None:
            # Name of the forecaster should be known only if the forecaster is null
            _err_msg = f'unsupported forecaster! Known forcasters are {list(CommonForecastingModels.keys())}'
            assert forecaster_name in CommonForecastingModels.keys(), _err_msg
            self.forecaster = CommonForecastingModels[forecaster_name]
        else:
            self.forecaster = forecaster

        # Forecasting parameters
        #1. horizon
        self.fh = ForecastingHorizon(np.arange(1, fh+1))
        #2. Initial training window
        self.initial_window = int(len(self.y)*pct_initial_window)
        #3. Step size
        self.step_length = step_length

        #4. Declare the cross validation window
        self.cv = ExpandingWindowSplitter(initial_window=self.initial_window,
                                          step_length=self.step_length,
                                          fh=self.fh)

        # Create the train test sets
        if self.X is None:
            self.X_train =  None
            self.X_test = None
            self.y_train, self.y_test = temporal_train_test_split(y=self.y, 
                                                                  train_size=self.initial_window)
        else:
            self.y_train, self.y_test, self.X_train, self.X_test = temporal_train_test_split(
            y=self.y, 
            X=self.X,
            train_size=self.initial_window)

        self.is_fitted = False
        self.fitted = None
        self.is_evaluated = False
        self.eval = None

        # Plots
        self.plot = self.__plot()

    def split_procedure_summary(self, verbose: bool=True) -> dict:
        """
        Generate a summary of the cross-validation procedure.

        Parameters:
        -----------        
            verbose : bool, optional
                If True, print the summary. Default is True.

        Returns:
        --------        
            dict
                A dictionary containing the summary of the cross-validation procedure.
        """

        _n_splits = self.cv.get_n_splits(self.y)
        cutoffs = [self.y.index[_train[-1]] for (_train, _) in self.cv.split(self.y.index)]
        _split_proc= {'Number of Folds': _n_splits,
                      'Initial Window Size': self.cv.initial_window,
                      'Step Length': self.cv.step_length,
                      'Forecast Horizon': len(self.cv.fh),
                      'First Cutoff': cutoffs[0],
                      'Last Curoff': cutoffs[-1]
                     }
        if verbose:
            for k, v in _split_proc.items():
                print(f"{k:<21}: {v}")
        return _split_proc    

    def __plot(self): #-> ForecastPlot
        """
        Create an instance of the ForecastPlot class for plotting utility.

        Returns:
        --------
            ForecastPlot:
                An instance of the ForecastPlot class.
        """
        return ForecastPlot(self)  

    def fit(self, 
            on: str = 'all', 
            fh: Optional[ForecastingHorizon] = None
           ): #-> ForecastFit
        """
        Fit the forecaster and compute insample results.

        Parameters:
        --------
            on : str, optional
                Either 'train' or 'all'. Sample on which the model is fitted. By default, it is fitted on the entire sample.
            fh : ForecastingHorizon, optional
                Forecast horizon.

        Returns:
        --------
            ForecastFit:
                An instance of the ForecastFit class containing the fitted model and insample performance metrics.
        """

        if fh is None:
            fh =  self.fh

        if on == 'all':
            self.forecaster.fit(y=self.y, X=self.X, fh=fh)
            self.is_fitted = True

        elif on=='train':
            self.forecaster.fit(y=self.y_train, X=self.X_train, fh=fh)
            self.is_fitted = False

        else: 
            on_values =['all', 'train']
            raise ValueError(f'argument takes 2 possible values {on_values}')
        self.fitted = ForecastFit(self)   
        return self.fitted

    def evaluate(self): #-> ForecastEval
        """
        Evaluate the forecaster out-of-sample.

        Returns:
        --------
            ForecastEval:
                An instance of the ForecastEval class containing the out-of-sample evaluation results.
        """

        self.eval = ForecastEval(self)  
        self.is_evaluated = True
        return self.eval

    def predict(self, 
                X: Optional[pd.DataFrame] = None, 
                fh: Optional[ForecastingHorizon] = None, 
                coverage: float = 0.9,
                verbose = False
               ) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """
        Generate predictions (Average and confidence interval) using the fitted model.

        Parameters:
        -----------        
            X : pd.DataFrame, optional
                Exogenous variables for forecasting. Default is None
            fh : ForecastingHorizon, optional
                Forecast horizon. Default is None and takes the horizon defined at instantiation.
            coverage : float, optional
                The coverage of the confidence interval. Default is 0.9.

        Returns:
        --------        
            Tuple[pd.DataFrame, pd.DataFrame]:
                A tuple containing the predictions and the confidence intervals.
        """
        if self.is_fitted==False:
            if verbose:
                print(f"\n{self.forecaster_name} model not fitted yet, or fitted on a subset only")
                print("Fitting the model on the entire sample ...")
            self.fit(on='all', fh=fh)

        if fh is None:
            fh =  self.fh
        if X is None:
            X = self.X

        y_pred = self.forecaster.predict(X=X, fh=fh)
        try:
            y_pred_ints = self.forecaster.predict_interval(X=X, fh=fh, coverage=coverage)
        except Exception as e:
            if verbose:
                print(f"{self.forecaster_name} does not support prediction intervals")
                print(f"Error: {e}")
                print("Computing the prediction intervals based on historical errors distribution")
            
            historical_errors = self.get_pred_errors()
            y_pred_ints = calculate_prediction_interval(historical_errors, y_pred, coverage=coverage)

        return y_pred, y_pred_ints

    def update(self, 
               new_y: pd.Series,
               new_X: Optional[pd.DataFrame] = None,
               fh: Optional[ForecastingHorizon] = None, 
               coverage: float = 0.9, 
               refit: bool = False,
              ) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """
        Update cutoff value to forecast new dates.
        Possibility to refit the model.

        Parameters:
        -----------        
            new_y : pd.Series
                The new dependent variable values.
            new_X : pd.DataFrame, optional
                The new exogenous variables. Default is None.
            fh : ForecastingHorizon, optional
                Forecast horizon. Default is None and takes the horizon defined at instantiation.
            coverage : float, optional
                The coverage of the confidence interval. Default is 0.9.
            refit : bool, optional
                If True, refit the model. Default is False.

        Returns:
        --------        
            Tuple[pd.DataFrame, pd.DataFrame]:
                A tuple containing the updated predictions and the updated confidence intervals.
        """
        new_y = new_y.resample(self.freq).last().ffill()
        if new_X is not None:
            new_X = new_X.resample(self.freq).last().ffill()
        self.forecaster.update(y=new_y, X=new_X, update_params=refit)
        y_pred, y_pred_ints = self.predict(X=new_X, fh=fh, coverage=coverage)
        return y_pred, y_pred_ints

    def get_pred_errors(self):
        """
        Get the prediction errors.
        Returns:
        --------        
            pd.DataFrame:
                A DataFrame containing the prediction errors.
        """
        if (self.is_evaluated is False) or (self.eval is None):
            self.eval = self.evaluate()        
        try:
            pred_errors = self.eval.oos_horizon_df[['cutoff', 'horizon', 'error']]
        except Exception as e:
            return None
        return pred_errors
        

    def __init_test(self,
                    data: pd.DataFrame, 
                    depvar_str: str, 
                    exog_l: Optional[list], 
                    freq: str
                   ) -> None:
        """
        Perform checks for the provided data, dependent variable, exogenous variables, and frequency.

        Parameters:
        -----------        
            data : pd.DataFrame
                A DataFrame containing the input data for forecasting.
            depvar_str : str
                The column name representing the dependent variable for forecasting.
            exog_l : list or None
                List of exogenous variables for forecasting.
            freq : str
                The frequency of the time series data.

        Returns:
        --------
            None

        Raises:
        -------        
            AssertionError:
                If the provided data is not a DataFrame, or if depvar_str is not a valid column in the data,
                or if exog_l is not None or an iterable.
            AssertionError:
                If any column in exog_l is not present in the data.
            AssertionError:
                If freq is not a recognized pandas frequency.
        """

        assert isinstance(data, pd.DataFrame), 'the data should be a dataframe'

        assert depvar_str in data.columns, 'the dependent varible must be a column in the data'
        self.depvar = depvar_str

        cond1 = exog_l is None
        cond2 = isinstance(exog_l, Iterable)
        assert cond1 or cond2, 'exog_l is either None, meaning no X or an iterable object'
        if cond2:
            assert all([c in data.columns for c in exog_l]), 'not all exog variables are in the data'
            self.exog_l = exog_l
        else:
            self.exog_l = exog_l

        assert freq in pandas_frequency_dict.keys(), f'Not a pandas recognized frequency. List of pandas frequencies:\n{pandas_frequency_dict}'
        self.freq = freq

        return None

    def __clean_data(self, 
                     data: pd.DataFrame, 
                     depvar_str: str, 
                     exog_l: Optional[list],
                     freq: str
                    ) -> Tuple[pd.Series, pd.DataFrame]:
        """
        Reformat the data taking into account the requested frequency.
        """

        _df = data.dropna().resample(freq).last().copy()
        # Declare and stage the variable to forecast
        y = _df[depvar_str]

        # List of Exogenous variables if any
        if exog_l is None:
            X = None
        else:
            assert all([c in _df.columns for c in exog_l]), 'not all columns are not in the data'
            X = _df[exog_l]
        return y, X

class ForecastPlot:
    """
    Plotting utility class for Forecast.

    Parameters:
    -----------    
        LF : Forecast
            An instance of the Forecast class
    """

    def __init__(self, LF: Forecast):
        self.y_train = LF.y_train
        self.y_test = LF.y_test
        self.y = LF.y
        self.cv = LF.cv
        return None

    def plot_train_test(self,
                        labels: List[str] = None,
                        xlabel: Optional[str] = None,
                        ylabel: Optional[str] = None,
                        title: str = 'Train-Test sets',
                        ax: Optional[plt.Axes] = None,
                        figsize: Tuple[float, float] = (15, 6)) -> None:
        """
        Plot the dependent variable separating the train from the test windows.

        Parameters:
        -----------
            labels : List[str], optional
                Labels for the plot. Default is ["y_train", "y_test"].
            xlabel : str, optional
                Label for the x-axis.
            ylabel : str, optional
                Label for the y-axis.
            title : str, optional
                The title of the plot. Default is 'Train-Test sets'.
            ax : plt.Axes, optional
                The Axes object for the plot.
            figsize : Tuple[float, float], optional
                The figure size. Default is (15, 6).

        Returns:
        --------        
            None
        """
        if labels is None:
            labels = ["y_train", "y_test"]
        return plot_series(self.y_train, self.y_test, 
                           labels =labels, 
                           ax = ax, 
                           xlabel = xlabel,
                           ylabel = ylabel,
                           title = title,
                           figsize = figsize)

    def plot_cv_procedure(self,
                          ax: Optional[plt.Axes] = None,
                          labels: List[str] = None,
                          ylabel: str = "Window number",
                          xlabel: str = "Time",
                          title: str = "Cross Validation Procedure"
                          ) -> None:
        """
        Plot the cross-validation procedure.

        Parameters:
        -----------        
            ax : plt.Axes, optional
                The Axes object for the plot.
            labels : List[str], optional
                Labels for the plot. Default is ["Window", "Forecasting horizon"].
            ylabel : str, optional
                Label for the y-axis. Default is "Window number".
            xlabel : str, optional
                Label for the x-axis. Default is "Time".
            title : str, optional
                The title of the plot. Default is "Cross Validation Procedure".

        Returns:
        --------        
            None
        """

        _train_windows, _test_windows = self._get_windows()
        if labels is None:
            labels = ["Window", "Forecasting horizon"]
            
        return plot_windows(self.y, 
                            _train_windows,
                            _test_windows,
                            ax = ax,
                            labels = labels,
                            xlabel = xlabel, 
                            ylabel = ylabel,
                            title = title)

    def plot_prediction(self,
                        y_pred: pd.Series,
                        y_pred_ints: Optional[pd.DataFrame] = None,
                        interval_label: str = 'CI',
                        labels: List[str] = None,
                        xlabel: Optional[str] = None,
                        ylabel: Optional[str] = None,
                        title: str = 'Prediction',
                        ax: Optional[plt.Axes] = None,
                        figsize: Tuple[float, float] = (15, 6)):
        """
        Plot the forecast predictions and the confidence intervals.

        Parameters:
        ----------        
            y_pred : pd.Series
                The predicted values.
            y_pred_ints : pd.DataFrame, optional
                The DataFrame containing the prediction intervals.
            interval_label : str, optional
                Label for the prediction interval. Default is 'prediction interval'.
            labels : List[str], optional
                Labels for the plot. Default is ["y_train", "y_pred"].
            xlabel : str, optional
                Label for the x-axis.
            ylabel : str, optional
                Label for the y-axis.
            title : str, optional
                The title of the plot. Default is 'Prediction'.
            ax : plt.Axes, optional
                The Axes object for the plot.
            figsize : Tuple[float, float], optional
                The figure size. Default is (15, 6).

        Returns:
        -------        
            None
        """

        y = self.y
        y_train = y.loc[y.index<y_pred.index[0]]
        zoom_y_train = y_train.iloc[-3*len(y_pred):]	
        if labels is None:
            labels = ["y_train", "y_pred"]
        return plot_series(zoom_y_train, y_pred,
                           labels= labels,
                           pred_interval=y_pred_ints,	
                           interval_label = interval_label,
                           xlabel = xlabel,
                           ylabel = ylabel,
                           title = title, 
                           ax = ax,
                           figsize = figsize
                          )

    def plot_prediction_true(self,
                             y_pred: pd.Series,
                             y_pred_ints: Optional[pd.DataFrame] = None,
                             interval_label: str = 'CI',
                             labels: List[str] = None,
                             xlabel: Optional[str] = None,
                             ylabel: Optional[str] = None,
                             title: str = 'Prediction',
                             ax: Optional[plt.Axes] = None,
                             figsize: Tuple[float, float] = (15, 6)):
        """
        Plot the forecast predictions, true values, and the confidence intervals.

        Parameters:
        -----------        
            y_pred : pd.Series
                The predicted values.
            y_pred_ints : pd.DataFrame, optional
                The DataFrame containing the prediction intervals.
            interval_label : str, optional
                Label for the prediction interval. Default is 'prediction interval'.
            labels : List[str], optional
                Labels for the plot. Default is ["y_train", "y_true", "y_pred"].
            xlabel : str, optional
                Label for the x-axis.
            ylabel : str, optional
                Label for the y-axis.
            title : str, optional
                The title of the plot. Default is 'Prediction'.
            ax : plt.Axes, optional
                The Axes object for the plot.
            figsize : Tuple[float, float], optional
                The figure size. Default is (15, 6).

        Returns:
        --------        
            None
        """

        y = self.y
        y_train = y.loc[y.index<y_pred.index[0]]
        zoom_y_train = y_train.iloc[-5*len(y_pred):]
        true_pred_idx = np.intersect1d(y.index, y_pred.index)
        err_msg = 'No overlap between true values and predicted values.\nIf you want to plot prediction alone use the function plot_prediction'
        assert len(true_pred_idx)>0, err_msg
        y_true = self.y[true_pred_idx]
        if labels is None:
            labels = ["y_train", "y_true", "y_pred"]

        return plot_series(zoom_y_train, y_true, y_pred,
                           pred_interval=y_pred_ints,
                           interval_label =interval_label,
                           labels=labels, 
                           xlabel = xlabel,
                           ylabel = ylabel,
                           title = title, 
                           ax = ax,
                           figsize = figsize)

    def _get_windows(self) -> Tuple[List[np.ndarray], List[np.ndarray]]:
        """
        Generate windows for the cross-validation procedure.

        Returns:
        --------
            tuple:
                A tuple containing two lists: train_windows and test_windows.
        """
        _train_windows = []
        _test_windows = []
        _y_index = self.y.index 
        for _train, _test in self.cv.split(_y_index):
            _train_windows.append(_train)
            _test_windows.append(_test)
        return _train_windows, _test_windows 

##############################################################################
# Model Fit and Insample Performance
##############################################################################
def compute_predictions(params):
    forecaster, X, intrain, intest, verbose = params
    fh = ForecastingHorizon(intest.index, is_relative=False)
    try:
        in_pred = forecaster.predict(fh=fh, X=X).rename('y_pred').reset_index()
        in_pred['y_true'] = intest.values
        in_pred['error'] = in_pred['y_true'] - in_pred['y_pred']
        in_pred['error_pct'] = in_pred['error'].abs()/in_pred['y_true']
        in_pred.insert(0, 'horizon', np.arange(1, len(in_pred) + 1))
        in_pred.insert(0, 'cutoff', intrain.index[-1])        
    except Exception as e:
        if verbose:
            print(f"Error occured in {intrain.index[-1]}: {e}")
        in_pred = pd.DataFrame()
    return in_pred
class ForecastFit:
    """
    Class for fitting the forecast model and computing insample performance metrics.

    Parameters:
    -----------    
        Forecast : Forecast
            An instance of the Forecast class.
    """

    def __init__(self, LF: Forecast):
        self.forecaster = LF.forecaster
        self.forecaster_name = LF.forecaster_name
        self.is_fitted = LF.is_fitted
        self.y_train = LF.y_train
        self.y = LF.y
        self.X = LF.X
        self.fh = LF.fh
        self.cv = LF.cv
        self.plot = self.__plot()

        self.insample_result_df = None
        self.insample_perf_summary = None

    def insample_predictions(self, random_sample=False, nsample: int = 100, verbose=False) -> pd.DataFrame:
        """
        Compute the insample predictions for the fitted model.

        Parameters:
        -----------        
            nsample : int, optional
                The number of samples to compute. Default is 100.

        Returns:
        --------        
            pd.DataFrame
                A DataFrame containing the insample predictions.
        """

        if self.is_fitted:
            _y = self.y
        else:
            _y = self.y_train

        insample_eval_window = len(_y) - len(self.fh)
        if random_sample:            
            nsample = max(insample_eval_window, nsample)
            # Randomly selecting cutoff points
            _cutoffs = np.random.choice(insample_eval_window, size=nsample, replace=False)
            cv_in = CutoffSplitter(cutoffs=_cutoffs, window_length=1, fh=self.fh)
        else:
            _cutoffs = np.arange(insample_eval_window)
            cv_in = ExpandingWindowSplitter(initial_window=1, step_length=1, fh=self.fh)
        if verbose:
            print(f"\nComputing {self.forecaster_name} forecaster historic predictions....")        

        params = [(self.forecaster, self.X, intrain, intest, verbose) for intrain, intest in cv_in.split_series(_y)]                
        with Pool() as pool:
            insample_result = list(tqdm(pool.imap(compute_predictions, params), total=len(params)))

        insample_result_df = pd.concat(insample_result)
        if insample_result_df.empty:
            print(f"No insample predictions computed {self.forecaster_name}")
        if verbose:
            print(f"\n{self.forecaster_name} forecaster historic predictions completed")        
        self.insample_result_df = insample_result_df
        return insample_result_df

    def insample_perf(self) -> dict:
        """
        Compute insample performance metrics (RMSE and MAPE) for the fitted model.

        Returns:
        --------        
            dict
                A dictionary containing the computed insample performance metrics.
        """

        if self.insample_result_df is None:
            self.insample_predictions()

        insample_perf_summary = {
            'cutoff': summary_perf(self.insample_result_df, grouper = 'cutoff', y_true_col = 'y_true', y_pred_col = 'y_pred'),
            'horizon':summary_perf(self.insample_result_df, grouper = 'horizon', y_true_col = 'y_true', y_pred_col = 'y_pred')
        }
        self.insample_perf_summary = insample_perf_summary
        return insample_perf_summary

    def __plot(self):
        return ForecastFitPlot(self)

class ForecastFitPlot:
    """
    Plotting utility class for ForecastFit.

    Parameters:
    -----------    
        LFF : ForecastFit
            An instance of the ForecastFit class.
    """

    def __init__(self, LFF: ForecastFit):
        self.LFF = LFF

    def plot_insample_performance(self,
                                  metric: str = 'RMSE',
                                  title: str = 'Insample Performance'):
        """
        Plot the insample performance metrics.

        Parameters:
        -----------        
            metric : str, optional
                The performance metric to plot. Default is 'RMSE'.
            title : str, optional
                The title of the plot. Default is 'Insample Performance'.

        Returns:
        --------        
            fig : plt.Figure
                The Figure object containing the plot.
            axes : np.array
                An array of Axes objects containing the plot.
        """

        assert metric in ['RMSE', 'MAPE'], f'{metric} not in summary performance'
        if 'insample_perf_summary' not in self.__dict__.keys():
            insample_perf_summary = self.LFF.insample_perf()

        f, axes = plt.subplots(1,2,figsize=(15,5))
        for i, (_grouper, _df) in enumerate(insample_perf_summary.items()):
            _df[metric].plot(ax= axes[i], style = '-o', title = f'{metric} By {_grouper}')
            axes[i].set_xlabel('')

        plt.suptitle(title)
        plt.tight_layout()                   
        return f, axes

##############################################################################
# Model Out of Sample Evaluation
##############################################################################
class ForecastEval:
    """
    Class for evaluating the forecaster out-of-sample.

    Parameters:
    -----------    
        Forecast : Forecast
            An instance of the Forecast class.
    """

    def __init__(self, LF: Forecast):
        
        self.forecaster = LF.forecaster
        self.forecaster_name = LF.forecaster_name
        self.y = LF.y
        self.X = LF.X
        self.cv = LF.cv

        self.scoring_metrics = [MeanSquaredError(square_root=True),
                                MeanAbsoluteError(),
                                MeanAbsolutePercentageError(), 
                                MedianAbsoluteError(),
                                ]

        _rename_metrics = {
            'test_MeanSquaredError':'RMSE', 
            'test_MeanAbsoluteError':'MAE',
            'test_MeanAbsolutePercentageError':'MAPE',
            'test_MedianAbsoluteError':'MedianAE',
        }
        print(f"\nStart {self.forecaster_name} forecaster evalution....")
        st = time.time()
        self.oos_eval = evaluate(forecaster=self.forecaster, 
                            y=self.y,
                            X = self.X,
                            cv=self.cv,
                            strategy="refit",
                            return_data=True,
                            scoring = self.scoring_metrics,
                            backend ='loky',
                           )
        self.oos_eval = self.oos_eval.set_index('cutoff').sort_index()
        self.oos_eval = self.oos_eval.rename(columns = _rename_metrics)
        et = time.time()
        elapsed_time = et - st
        print(f"Evaluation completed in: {np.around(elapsed_time / 60,3)} minutes")

        convert_horizon = self.oos_eval.apply(self.__eval_horizon, axis=1)
        
        self.oos_horizon_df = pd.concat(convert_horizon.values)
        self.oos_horizon_perf = summary_perf(self.oos_horizon_df, 
                                             grouper='horizon', 
                                             y_true_col = 'y_test', 
                                             y_pred_col = 'y_pred')
        self.oos_cutoff_perf = summary_perf(self.oos_horizon_df, 
                                             grouper='cutoff', 
                                             y_true_col = 'y_test', 
                                             y_pred_col = 'y_pred')

        self.plot = self.__plot()
        return None

    def summary_results(self) -> pd.DataFrame:
        """
        Generate a summary of out-of-sample forecast results.

        Returns:
        --------        
            pd.DataFrame
                A DataFrame containing various summary statistics of the out-of-sample forecasts.
        """

        _summary = {
         'Number of Folds':  self.oos_eval.shape[0], 
         'Avg Fit time (s)': self.oos_eval.fit_time.mean(),
         'Avg_pred_time (s)': self.oos_eval.pred_time.mean(),
         'Smallest training window': self.oos_eval.len_train_window.min(),
         'Largest training window':self.oos_eval.len_train_window.max(),
         'First cutoff': self.oos_eval.index[0],
         'Last cutoff': self.oos_eval.index[-1],
        }
        for _s in self.oos_cutoff_perf.columns:
            _summary[f'Avg {_s}'] = self.oos_cutoff_perf[_s].mean()
        return pd.Series(_summary).to_frame().T

    def summary_cutoff(self) -> pd.DataFrame:
        """
        Generate a summary of out-of-sample performance per cutoff.

        Returns:
        --------        
            pd.DataFrame
                A DataFrame containing summary performance metrics (RMSE and MAPE) for each horizon.
        """        
        return self.oos_cutoff_perf

    def summary_horizon(self) -> pd.DataFrame:
        """
        Generate a summary of out-of-sample performance per horizon.

        Returns:
        --------        
            pd.DataFrame
                A DataFrame containing summary performance metrics (RMSE and MAPE) for each horizon.
        """        
        return self.oos_horizon_perf

    def __eval_horizon(self, x):
        _fct = pd.concat([x['y_test'], x['y_pred']], keys=['y_test', 'y_pred'], axis=1)
        _fct['error'] = _fct['y_test'] - _fct['y_pred']
        _fct['error_pct'] = _fct['error'].abs()/ _fct['y_test']
        _fct['horizon'] = np.arange(1, len(_fct)+1)    
        _fct['cutoff'] = x.name
        return _fct

    def __plot(self):
        return ForecastEvalPlot(self)

class ForecastEvalPlot:
    """
    Plotting utility class for ForecastEval.

    Parameters:
    -----------    
        LFE : ForecastEval
            An instance of the ForecastEval class.
    """

    def __init__(self, LFE: ForecastEval):
        self.oos_horizon_perf = LFE.oos_horizon_perf
        self.oos_cutoff_perf = LFE.oos_cutoff_perf
        return None

    def plot_oos_score(self,
                       score: str = 'RMSE',
                       view: str = 'horizon', 
                       xlabel: str = None,
                       ylabel: str = None,
                       title: str = 'Out of Sample Performance',
                       ax: Optional[plt.Axes] = None,
                       figsize: Tuple[float, float] = (15, 6)):
        """
        Plot out-of-sample performance metric historically.

        Parameters:
        -----------        
            score : str, optional
                The performance metric to plot. Default is 'RMSE'.
            view: str, optional
                The view of the plot. It must be either 
                    - horizon: the function plots the average score per forecast horizon
                    - cutoff: the function plots the average across all horizons for each cutoff                
                Default is 'horizon'.
            xlabel : str, optional
                Label for the x-axis.
            ylabel : str, optional
                Label for the y-axis.
            title : str, optional
                The title of the plot. Default is 'Out of Sample Performance'.
            ax : plt.Axes, optional
                The Axes object for the plot.
            figsize : Tuple[float, float], optional
                The figure size. Default is (15, 6).
        
        Returns:
        --------        
            fig : plt.Figure
                The Figure object containing the plot.
            axes : np.array
                An array of Axes objects containing the plot.

        """
        if view == 'horizon':
            assert score in self.oos_horizon_perf.columns, 'score not computed'
            to_plot = self.oos_horizon_perf[score]
        elif view == 'cutoff':
            assert score in self.oos_cutoff_perf.columns, 'score not computed'
            to_plot = self.oos_cutoff_perf[score]
        else:
            raise ValueError('view should be either horizon or cutoff')

        if ax is None:
            f, ax = plt.subplots(1,1,figsize=figsize)

        ylabel = score if ylabel is None else ylabel
        xlabel = '' if xlabel is None else xlabel        
        to_plot.plot(ax = ax, style = '-o')                    
        ax.set(xlabel = xlabel, ylabel=ylabel)
        ax.set_title(title, size ='xx-large')
        if ax is None:
            return f, ax
        else:
            return ax


import pandas as pd
import numpy as np
from scipy.stats import norm
from sklearn.metrics import (
    mean_squared_error,
    mean_absolute_error,
    r2_score,
    median_absolute_error
)

def evaluate_metrics(y_true, y_pred):
    """
    Computes performance metrics for the fitted curve.
    The performance metrics include 
        RMSE: Root Mean Squared Error
        MAE: Mean Absolute Error
        MAPE: Mean Absolute Percentage Error
        R2: R-squared
        MedianAE: Median Absolute Error
    """
    metrics = {
        'RMSE': np.sqrt(mean_squared_error(y_true, y_pred)),
        'MAE': mean_absolute_error(y_true, y_pred),
        'MAPE': np.mean(np.abs((y_true - y_pred) / y_true)),
        'R2': r2_score(y_true, y_pred),
        'MedianAE': median_absolute_error(y_true, y_pred)
    }
    return pd.Series(metrics)

def summary_perf(insample_result_df, 
			 grouper, 
			 y_true_col = 'y_true',
			 y_pred_col = 'y_pred',
			):
    """
	Compute summary performance metrics for a given forecast.
        RMSE: Root Mean Squared Error
        MAE: Mean Absolute Error
        MAPE: Mean Absolute Percentage Error
        R2: R-squared
        MedianAE: Median Absolute Error

	Parameters:
    -----------
	    insample_result_df (DataFrame): 
            A DataFrame containing the forecast results.
		grouper (str): 
            The column name to group the forecast results for computing summary metrics.
		y_true_col (str, optional):
            The column name representing the true values. Default is 'y_true'.
		y_pred_col (str, optional): 
            The column name representing the predicted values. Default is 'y_pred'.

	Returns:
    --------
		DataFrame: A DataFrame containing summary performance metrics.
	"""

    horizon_metrics = insample_result_df.groupby(grouper).apply(lambda x: evaluate_metrics(x[y_true_col], x[y_pred_col]))
    return horizon_metrics

def calculate_prediction_interval(historical_errors, y_pred, coverage=0.90):
    """
    Calculate the prediction interval for given predictions based on historical errors.

    Parameters:
    -----------
    historical_errors : array-like
        Array of historical prediction errors (actual - predicted values).
    y_pred : float or array-like
        Predicted value(s) for which the prediction interval is required.
    coverage : float, optional
        Desired coverage of the prediction interval. Default is 0.90.

    Returns:
    --------
    interval : tuple
        A tuple containing the lower and upper bounds of the prediction interval.
    """
    # Calculate the mean and standard deviation of the errors
    sigma_e = historical_errors.groupby('horizon').error.std().sort_index()
    sigma_e.index = y_pred.index

    # Determine the z-score for the desired coverage
    z = norm.ppf((1 + coverage) / 2)

    # Calculate the prediction interval
    lower_bound = y_pred - z * sigma_e
    upper_bound = y_pred + z * sigma_e

    ypred_int = pd.concat([lower_bound, upper_bound], axis=1)
    ypred_int.columns = pd.MultiIndex.from_tuples([(y_pred.name, coverage, 'lower'),
                                                  (y_pred.name, coverage, 'upper')])
    return ypred_int



""" Combination of timeseries Forecasts"""
__description__ = "Time Series Forecast Combination"
__author__ = "Amine Raboun - amineraboun@github.io"

import logging
logging.getLogger().setLevel(logging.ERROR)
import warnings
warnings.filterwarnings("ignore")
warnings.simplefilter("ignore", category=FutureWarning)
warnings.simplefilter("ignore", category=DeprecationWarning)

##############################################################################
# Import Libraries & default configuration 
##############################################################################
import pandas as pd
import numpy as np
from typing import Union, Optional, Tuple, List, Dict, Any

from .forecast import Forecast
from .forecast import CommonForecastingModels
from .utils.plotting import plot_series
from sktime.forecasting.base import ForecastingHorizon

import matplotlib.pyplot as plt
plt.rc('axes', titlesize='x-large')    
plt.rc('axes', labelsize='large')   
plt.rc('xtick', labelsize='large')   
plt.rc('ytick', labelsize='large')   
plt.rc('legend', fontsize='large')   
plt.rc('figure', titlesize='x-large') 
import seaborn as sns
sns.set_theme(style='white', font_scale=1)

from concurrent.futures import ProcessPoolExecutor
def fetch_errors(item):
    _fname, _lf = item
    return _fname, _lf.get_pred_errors()


class ForecastModelSelect:
    """
    Class for model selection and comparison based on out-of-sample performance.
    Evaluate how to best combine the different models based on their oos performance.

    The class can be initialized in 2 ways. 
        Method 1: Requires a list of Forecast objects. It is the preferable instantiation if the models have already been evaluated out of sample.

        Method 2: Requires all the arguments taken by Forecast plus a dictionary where the keys are names of forecasts and the values are the forecast_models.

    Parameters:
    -----------    
        forecasters_d : dict, optional
            A dictionary containing various forecasting models for comparison. 
            Default is None and assess the most common forecasting models.
                Naive: NaiveForecaster - Keep the latest value
                AutoARIMA: StatsForecastAutoARIMA - Auto ARIMA model
                AutoETS: StatsForecastAutoETS - Auto ETS model
                AutoTheta: StatsForecastAutoTheta - Auto Theta model
                TBATS: StatsForecastAutoTBATS - TBATS model
                LOESS: StatsForecastMSTL - LOESS model
                Prophet: Prophet - Prophet model
        trained_forecasters_d : dict, optional
            A dictionary containing trained Forecast objects. Default is None.
        mode : str, optional
            The aggregation mode. Default is 'nbest_average_horizon'.
        score : str, optional
            The performance score. Default is 'RMSE'.
        nbest : int, optional
            Number of best models to aggregate. Default is 2.
        kwargs
            Additional keyword arguments to be passed to the Forecast initialization.

    Raises:
    -------    
        AssertionError
            If the mode, score, or nbest values are not recognized or do not meet the requirements.
            If both LF_list and forecasters_d are None.
    """

    def __init__(self, 
                 forecasters_d: Optional[Dict] = None,
                 trained_forecasters_d: Optional[Dict] = None,
                 model_exog_d: Optional[Dict] = None,
                 mode: str = 'best_horizon',
                 score: str = 'RMSE',
                 nbest: int = None,
                 **kwargs: Any
                ) -> None:

        lf_d = {}
        if (trained_forecasters_d is not None):
            # Trained Models must be istance of Forecast object
            assert len(set(forecasters_d.keys()).intersection(set(trained_forecasters_d.keys()))) == 0, 'There is an overlap between the two trained and non trainded dictionaries' 
            if (forecasters_d is not None):
                # There must be no overlap between the two dictionaries
                assert all([isinstance(_lf, Forecast) for _lf in trained_forecasters_d.values()]), 'Trained models must be instances of Forecast object'
            for forecaster_name, _lf in trained_forecasters_d.items():
                lf_d[forecaster_name] = _lf        
        else:
            if (forecasters_d is None):
                forecasters_d = CommonForecastingModels
        
        # Define the Forecasters exogenous variables
        exog_l = kwargs.pop('exog_l', None)
        if model_exog_d is None:
            model_exog_d = {forecaster_name: exog_l for forecaster_name in forecasters_d.keys()}
        else:
            exog_l = [] if exog_l is None else exog_l
            for forecaster_name in forecasters_d.keys():
                model_exog_d[forecaster_name] = exog_l + model_exog_d.get(forecaster_name, [])
        
        if forecasters_d is not None:
            for forecaster_name, forecaster in forecasters_d.items():            
                _lf =  Forecast(
                    forecaster_name = forecaster_name,
                    forecaster = forecaster,
                    exog_l = model_exog_d[forecaster_name],
                    **kwargs
                )
                lf_d[forecaster_name] = _lf
                           
        self.LF_d = lf_d

        assert isinstance(mode, str), 'mode must be a sting'
        recog_modes = ['best', 'best_horizon', 'average', 'inverse_score', 
                           'nbest_average', 'nbest_average_horizon']
        assert mode in recog_modes, f'mode function not implemented!. Recognized modes are {recog_modes}'
        self.mode = mode

        recog_scores = ['RMSE', 'MAE', 'MAPE', 'R2', 'MedianAE']
        assert score in recog_scores, f'performance score not implemented!. Recognized scores are {recog_scores}'		
        self.score = score

        if nbest is None:
            nbest = 2
        else:
            assert isinstance(nbest, int), 'n best must be an integer'		
        self.nbest = min(nbest, len(self.LF_d))	

        self.eval_models = None
        self.summary_horizon = None
        self.summary_results = None
        self.best_x_overall = None
        self.model_rank_perhorizon = None
        self.avg_oos_horizon = None
        self.avg_oos_hist = None
        return None

    def split_procedure_summary(self):
        """
        Print the summary of the split procedure for each model.

        Returns:
        --------        
            dict:
                A dictionary containing details of the cross-validation procedure, including:
                the number of folds, initial window size, step length, and forecast period.
        """
        _lf  = list(self.LF_d.values())[0]
        return _lf.split_procedure_summary()
    
    def add_forecaster(self, 
                       forecaster_name: str,
                       lf: Forecast) -> None:
        """
        Add a Forecast Object to the list of models to evaluate.

        Parameters:
        -----------
            lf : Forecast
                The Forecast object to be added.
        """
        assert forecaster_name not in self.LF_d.keys(), 'Forecaster name already exists in the considered list of model'
        self.LF_d[forecaster_name] = lf
        return self.LF_d

    def fit(self, 
            on: str = 'all', 
            fh: Optional[ForecastingHorizon] = None, 
            force: bool = False,
            verbose: bool = False
           ) -> None:
        """
        Fit the forecasting models for all the underlying Forecast objects.

        Parameters:
        -----------        
            on : str, optional
                The period for the in-sample fitting. Default is 'all'.
            fh : ForecastingHorizon, optional
                The forecasting horizon. Default is None and it takes the value entered at initialization..
            force : bool, optional
                If True, force fitting even if the models are already fitted. Default is False.

        Returns:
        --------        
            None
        """
        for _lfname, _lf in self.LF_d.items():
            if verbose:
                print(f'Fitting {_lfname} ...')
            if (_lf.is_fitted) and force ==False:
                pass
            else:
                _lf.fit(on=on, fh=fh)
        return None

    def evaluate(self,
                 force: bool = False
                ) -> Tuple:
        """
        Evaluate the underlying Forecast models out of sample.

        Parameters:
        -----------
            force : bool, optional
                If True, force the evaluation even if the models are already evaluated. Default is False.

        Returns:
        --------
            tuple
                A tuple containing the summary of out-of-sample performance per horizon and per model.
        """
        #Step 1: Loop on all models and evaluate them
        _model_evals = []
        _todrop_evals = []
        for _fname, _lf in self.LF_d.items():
            if (_lf.is_evaluated) and force ==False:
                _model_evals.append(_lf.eval)
            else:
                try:
                    _eval = _lf.evaluate()
                    _model_evals.append(_eval)
                except Exception as e:
                    print(f'Error evaluating {_fname}: {e}')
                    print(f'\nmodel {_fname} cannot be evaluated. It will be removed from the list of models')
                    _todrop_evals.append(_fname)
                    continue                
        if len(_todrop_evals)>0:
            for _fname in _todrop_evals:
                self.LF_d.pop(_fname)
        self.eval_models = _model_evals

        # Step 2: Get the OOS Summary of Performance 
        #per horizon and per model
        _sum_h = {}; _sum_T = {}
        for _lf_eval in self.eval_models:            
            _sum_T[_lf_eval.forecaster_name] = _lf_eval.summary_results().squeeze()
            _sum_h[_lf_eval.forecaster_name] = _lf_eval.summary_horizon()

        self.summary_horizon = pd.concat(_sum_h.values(), axis=1, keys = _sum_h.keys())
        self.summary_results = pd.concat(_sum_T.values(), axis=1, keys = _sum_T.keys())
        return self.summary_horizon, self.summary_results

    def select_best(self,
                    score: Optional[str] = None,
                    reestimate = False,
                   ) -> pd.DataFrame:
        """
        Select the best model based on horizon and overall performance.

        Parameters:
        -----------        
            score : str, optional
                The performance metric to use for model comparison. Default is None and it takes the value entered at initialization.
            nbest : int, optional
                The number of best models to select based on horizon performance. Default is None and it takes the value entered at initialization.

        Returns:
        --------        
            pd.DataFrame
                model_rank_perhorizon. Rank of models per horizon based on performance.
        """
        if score is None:
            score = self.score

        if (self.model_rank_perhorizon is None) or (self.avg_oos_horizon is None) or reestimate:

            # if models are not evaluated yet, run evaluate
            cond1 = self.eval_models is None
            cond2 = self.summary_horizon is None
            if cond1 or cond2:	
                print('\nRun evaluate ...')
                self.evaluate()

            nevals = len(self.LF_d)
            nbest = self.nbest
            if nbest > nevals:
                print(f'\nnbest ={nbest} is higher than the number of models evaluated {nevals}')
                print('All models will be considered. The average best models will be equal to the simple average')
                nbest = max(nbest, nevals)

            best_x_overall = {}
            model_rank_perhorizon = {}
            avg_oos_horizon = {}
            _perf_metrics = list(self.summary_horizon.columns.levels[1])
            for _s in _perf_metrics:
                r = self.summary_horizon.unstack().unstack(1)[_s].unstack()

                # Select best x models based on their average performance on all horizons 
                overall_score = r.mean(axis=1).sort_values(ascending=True)
                best_overall = overall_score.index[0]
                _s_best_x_overall = list(overall_score.index[:nbest])

                # Summary Model average
                _avg_oos_horizon = pd.concat([r.loc[best_overall],
                                            r.min(),
                                            r.loc[_s_best_x_overall].mean(),
                                            r.apply(lambda x: x.nsmallest(nbest).mean(), axis=0),
                                            r.mean(), 
                                            ], 
                                            keys = [
                                            'Best Model (over all)',
                                            'Best Model (per horizon)',
                                            f'Best {nbest} Models (over all)',
                                            f'Best {nbest} Models (per horizon)',
                                            'Model Avg (all models)'
                                            ], 
                                            axis = 1)

                avg_oos_horizon[_s] = pd.concat([r.T, _avg_oos_horizon], axis=1)

                # Compute the rank of the models per horizon
                _rank_perhorizon = r.apply(lambda x: x.nsmallest(nevals).index.tolist(), axis=0)\
                    .reset_index(drop=True).rename(index={i: f'Best_{i+1}' for i in range(nevals)})
                model_rank_perhorizon[_s] = _rank_perhorizon
                best_x_overall[_s] = _s_best_x_overall
            
            self.best_x_overall = best_x_overall
            self.model_rank_perhorizon = model_rank_perhorizon
            self.avg_oos_horizon = avg_oos_horizon

        return self.model_rank_perhorizon[score], self.avg_oos_horizon[score]

    def oos_per_cutoff(self,
                       score: Optional[str] = None
                      ) -> pd.DataFrame:
        """
        Calculate and return the out-of-sample performance based on cutoffs.

        Parameters:
        -----------        
            score : str, optional
                The performance metric to use for model comparison. Should be either 'RMSE' or 'MAPE'. Default is None and it takes the value entered at initialization.
            nbest : int, optional
                The number of best models to select based on horizon performance. Default is None and it takes the value entered at initialization.

        Returns:
        --------        
            pd.DataFrame
                A DataFrame containing the out-of-sample performance per horizon and per model based on the specified score and nbest values.
        """    
        if score is None:
            score = self.score
        
        if self.avg_oos_hist is None:
            nbest = self.nbest
            nevals = len(self.LF_d)
            if nbest > nevals:
                print(f'\nnbest ={nbest} is higher than the number of models evaluated {nevals}')
                print(f'The average performance of best models will be computed on {nevals} models instead')
                nbest = max(nbest, nevals)

            if self.eval_models is None:
                print('Run evaluate ...')
                self.evaluate()

            avg_oos_hist = {}
            _perf_metrics = list(self.summary_horizon.columns.levels[1])
            for _s in _perf_metrics:
                _oos = {_lf_eval.forecaster_name: _lf_eval.oos_cutoff_perf[_s] for _lf_eval in self.eval_models}
                _oos = pd.concat(_oos.values(),  axis=1, keys= _oos.keys())

                best_mod = _oos.mean().idxmin()
                nbest_mods = list(_oos.mean().nsmallest(nbest).index)
                _aggs = pd.concat([_oos[best_mod], _oos[nbest_mods].mean(axis=1), _oos.mean(axis=1)], axis = 1,
                                  keys = ['Best Model (over all)', f'Best {nbest} Models (over all)','Model Avg (all models)'])
                avg_oos_hist[_s] = pd.concat([_oos, _aggs], axis=1)
            self.avg_oos_hist = avg_oos_hist
        
        return avg_oos_hist[score]

    def __stage_X(self,
                  LF:Forecast,
                  X: pd.DataFrame
                  ) -> None:
        if LF.X is not None:
            _lf_exogs = LF.X.columns
            assert all([_ex in X.columns for _ex in _lf_exogs]), f'Some exogenous variables are missing for model {LF.forecaster_name}'
            _lf_X = X[_lf_exogs]
        else:
            _lf_X = None
        return _lf_X
        
    def predict(self,
                X: Optional[pd.DataFrame] = None, 
                fh: Optional[ForecastingHorizon] = None,
                coverage: float = 0.9,
                mode: Optional[str] = None,
                score: Optional[str] = None,
                model_name = None,
                ret_underlying: bool = False
               ) -> Tuple[pd.Series, pd.DataFrame]:
        """
        Make forecasts using the specified aggregation mode.

        Parameters:
        -----------        
            X : pd.DataFrame, optional
                The exogenous variables used for prediction. Default is None and it takes the value entered at initialization..
            fh : ForecastingHorizon, optional
                The forecasting horizon. Default is None and it takes the value entered at initialization..
            coverage : float, optional
                The prediction interval coverage. Default is 0.9.
            mode : str, optional
                The aggregation mode for predictions. Default is None and it takes the value entered at initialization.. 
                Available values:
                * 'best': The prediction is based on the best model.
                * 'best_horizon': The prediction is based on the best model for each horizon.
                * 'average': The average of the prediction of all models.
                * 'inverse_score': The weighted average prediction, where weights are inversely proportional to the model performance score.
                * 'nbest_average': Average of the n best models. The n is given by the parameter nbest.
                * 'nbest_average_horizon': Average of the n best models for each horizon. The n is given by the parameter nbest.
                * 'model': Returns the prediction of a specific model. The model name must be provided in the model_name parameter.
            score : str, optional
                The performance metric to use for model comparison. Should be either 'RMSE' or 'MAPE'. Default is None and it takes the value entered at initialization..
            nbest : int, optional
                The number of best models to select based on horizon performance. Default is None and it takes the value entered at initialization..
            model_name: str, optional
                The model name to use for prediction if mode='model'. Default is None.
            ret_underlying : bool, optional
                If True, return the underlying predictions and prediction intervals for each model. Default is False.

        Returns:
        --------        
            tuple
                A tuple containing the aggregated prediction and prediction intervals.
        """
        if mode is None:
            mode = self.mode
        if mode == 'model':
            assert model_name in self.LF_d.keys(), 'Model name not in the list of models'
            _lf_X = self.__stage_X(self.LF_d[model_name], X)
            return self.LF_d[model_name].predict(X=_lf_X, fh=fh, coverage=coverage)
        else:
            preds = {}; pred_ints = {}
            for _fname, _lf in self.LF_d.items():
                _lf_X = self.__stage_X(self.LF_d[_fname], X)                
                _y_pred, _y_pred_ints = _lf.predict(X=_lf_X, fh=fh, coverage=coverage)
                preds[_fname] = _y_pred
                pred_ints[_fname] =_y_pred_ints

            preds = pd.concat(preds.values(), keys = preds.keys(), axis=1)
            pred_ints = pd.concat(pred_ints.values(), keys = pred_ints.keys(), axis=1)

            return self.__aggregate_pred(mode=mode, preds=preds, pred_ints=pred_ints, score=score, ret_underlying=ret_underlying)

    def update(self,
               new_y: pd.Series,
               new_X: Optional[pd.DataFrame] = None,
               refit: bool = False, 
               reevaluate:bool = False, 
               fh: Optional[ForecastingHorizon] = None, 
               coverage: float = 0.9, 
               mode: Optional[str] = None,
               score: Optional[str] = None,
               model_name = None,
               ret_underlying: bool = False
              ) -> Tuple[pd.Series, pd.DataFrame]:
        """
        Function to update the prediction for all the models and aggregate them based on the specific mode.

        Parameters:
        -----------        
            newdata : pd.DataFrame
                The new data for updating the predictions.
            refit : bool, optional
                If False, generate the prediction without fitting the model. If True, fits the underlying models again on the entire sample
                Default is False
            reevaluate: bool, optional    
                Reapply the cross-validateion, evaluate the models out of sample, select the best models, then aggregate the predictions according the setup.
                Default is False.
            fh : ForecastingHorizon, optional
                The forecasting horizon. Default is None and it takes the value entered at initialization..
            coverage : float, optional
                The prediction interval coverage. Default is 0.9.
            mode : str, optional
                The aggregation mode for predictions. Default is None and it takes the value entered at initialization..
                Available values:
                * 'best': The prediction is based on the best model.
                * 'best_horizon': The prediction is based on the best model for each horizon.
                * 'average': The average of the prediction of all models.
                * 'inverse_score': The weighted average prediction, where weights are inversely proportional to the model performance score.
                * 'nbest_average': Average of the n best models. The n is given by the parameter nbest.
                * 'nbest_average_horizon': Average of the n best models for each horizon. The n is given by the parameter nbest.
                * 'model': Returns the prediction of a specific model. The model name must be provided in the model_name parameter.
            score : str, optional
                The performance metric to use for model comparison. Should be either 'RMSE' or 'MAPE'. Default is None and it takes the value entered at initialization..
            nbest : int, optional
                The number of best models to select based on horizon performance. Default is None and it takes the value entered at initialization. and it takes the value entered at initialization..
            model_name: str, optional
                The model name to use for prediction if mode='model'. Default is None.
            ret_underlying : bool, optional
                If True, return the underlying predictions and prediction intervals for each model. Default is False.

        Returns:
        -------
            tuple
                A tuple containing the aggregated prediction and prediction intervals.
        """
        if mode is None:
            mode = self.mode

        if mode == 'model':
            assert model_name in self.LF_d.keys(), 'Model name not in the list of models'
            _lf_X = self.__stage_X(self.LF_d[model_name], new_X)
            return self.LF_d[model_name].update(new_y=new_y, new_X=_lf_X, fh=fh, coverage=coverage, refit=refit)
        else:
            if reevaluate == False:            
                preds = {}; pred_ints = {}
                for _lf in self.LF_d.values():
                    _lf_X = self.__stage_X(_lf, new_X)
                    try:
                        _y_pred, _y_pred_ints = _lf.update(new_y=new_y, new_X=_lf_X, 
                                                           fh=fh, coverage=coverage, refit=refit)
                        preds[_lf.forecaster_name] = _y_pred
                        pred_ints[_lf.forecaster_name] =_y_pred_ints
                    except Exception as e:
                        print(f'Error updating {_lf.forecaster_name}: {e}')
                        print(f"model {_lf.forecaster_name} cannot be updated. It won't be considered for forecasts")
                        continue                   

                preds = pd.concat(preds.values(), keys = preds.keys(), axis=1)
                pred_ints = pd.concat(pred_ints.values(), keys = pred_ints.keys(), axis=1)
                return self.__aggregate_pred(mode=mode, preds=preds, pred_ints=pred_ints, score=score, ret_underlying=ret_underlying)
            else:
                self.evaluate(force=True)
                self.select_best(score = score, reestimate=True)
                preds = {}; pred_ints = {}
                for _lf in self.LF_d.values():
                    _lf_X = self.__stage_X(_lf, new_X)
                    try:
                        _y_pred, _y_pred_ints = _lf.update(new_y=new_y, new_X=_lf_X, fh=fh, coverage=coverage, refit=refit)
                        preds[_lf.forecaster_name] = _y_pred
                        pred_ints[_lf.forecaster_name] =_y_pred_ints
                    except Exception as e:
                        print(f'Error updating {_lf.forecaster_name}: {e}')
                        print(f'model {_lf.forecaster_name} cannot be updated. It will be considered for forecasts')
                        continue                   
                preds = pd.concat(preds.values(), keys = preds.keys(), axis=1)
                pred_ints = pd.concat(pred_ints.values(), keys = pred_ints.keys(), axis=1)
                return self.__aggregate_pred(mode=mode, preds=preds, pred_ints=pred_ints, score=score, ret_underlying=ret_underlying)

    def save(self, path: str):
        """
        Save the model to a file.
        Parameters:
        -----------
            path : str
                The path to save the model.
        Returns:
        --------
            None
        """
        import pickle
        with open(path, 'wb') as f:
            pickle.dump(self, f)
        return None
    
    def get_pred_errors(self, 
                        mode: Optional[str] = None,
                        score: Optional[str] = None,
                        model_name = None, 
                        ):
        """
        Get the prediction errors.
        """
        if mode is None:
            mode = self.mode
        if mode == 'model':
            assert model_name in self.LF_d.keys(), 'Model name not in the list of models'
            return self.LF_d[model_name].get_pred_errors()
        else:
            if score is None:
                score = self.score
            
            # Using ProcessPoolExecutor to execute computation-heavy tasks in parallel
            with ProcessPoolExecutor() as executor:
                results = executor.map(fetch_errors, self.LF_d.items())
            errors = {fname: err for fname, err in results}
            for _fname, _lf in self.LF_d.items():
                _lf.fitted.insample_result_df = errors[_fname]

            # Convert results to dictionary            
            errors_df = pd.concat(errors.values(), keys = errors.keys(), axis=0).reset_index()
            errors_df = errors_df.rename(columns = {'level_0': 'forecaster'})


            if mode =='best':
                return errors[self.best_x_overall[score][0]].reset_index(drop=True)
            
            elif mode =='best_horizon':
                if self.model_rank_perhorizon is None:
                    self.select_best()
                best_horizon = self.model_rank_perhorizon[score].loc['Best_1'].to_dict()
                error_horizon_l = [errors[best_mod_h].loc[errors[best_mod_h]['horizon']==h] for h, best_mod_h in best_horizon.items()]
                return pd.concat(error_horizon_l, axis=0).reset_index(drop=True)
            
            elif mode =='average':
                return errors_df.groupby(['cutoff', 'horizon']).error.mean().reset_index()               
            
            elif mode =='inverse_score':
                if self.summary_results is None:
                    self.evaluate()
                _score = self.summary_results.loc[f'Avg {score}']
                _score = _score/_score.sum()
                errors_df['_score'] = errors_df['forecaster'].map(_score.to_dict())
                errors_df['weighted_error'] = errors_df['_score']*errors_df['error']
                return errors_df.groupby(['cutoff', 'horizon']).weighted_error.sum().reset_index()
            
            elif mode =='nbest_average':
                if self.best_x_overall is None:
                    self.select_best()
                return errors_df.loc[errors_df.forecaster.isin(self.best_x_overall[score])].groupby(['cutoff', 'horizon']).error.mean().reset_index()
            
            elif mode =='nbest_average_horizon':
                if self.model_rank_perhorizon is None:
                    self.select_best()
                best_horizon = self.model_rank_perhorizon[score].iloc[:self.nbest].T
                best_horizon = best_horizon.apply(lambda x: x.values, axis=1).to_dict()
                errors_best_horizon = []
                for h, best_mod_h in best_horizon.items():
                    errors_best_horizon.append(errors_df.loc[(errors_df.horizon==h) & errors_df.forecaster.isin(best_mod_h)]\
                        .groupby(['cutoff', 'horizon']).error.mean().reset_index())
                return pd.concat(errors_best_horizon, axis=0).reset_index(drop=True)            
            
            else:
                raise ValueError('mode can take the foloowing values: best, best_horizon, average, inverse_score, nbest_average, nbest_average_horizon, model')        
    
    def plot_model_compare(self, 
                           score: str = 'RMSE', 
                           view: str = 'horizon', 
                           model_subset: Optional[List[str]] = None,
                           xlabel: Optional[str] = None, 
                           ylabel: Optional[str] = None,
                           title: str = 'Out of Sample Performance', 
                           ax: Optional[plt.Axes] = None, 
                           figsize: Tuple[int, int] = (15, 6)
                          ) -> Union[plt.Figure, plt.Axes]:
        """
        Plot a comparison of models based on their out-of-sample performance.

        Parameters:
        -----------        
            score : str, optional
                The performance metric for comparison. Should be either 'RMSE' or 'MAPE'. Default is 'RMSE'.
            view : str, optional
                The view mode for comparison, either 'horizon' or 'cutoff'. Default is 'horizon'.
            xlabel : str, optional
                The label for the x-axis. Default is an empty string.
            ylabel : str, optional
                The label for the y-axis. Default is an empty string.
            title : str, optional
                The title of the plot. Default is 'Out of Sample Performance'.
            ax : matplotlib.axes._subplots.AxesSubplot, optional
                The matplotlib axes to use for plotting. Default is None.
            figsize : tuple, optional
                The size of the figure in inches (width, height). Default is (15, 6).

        Returns:
        --------        
            matplotlib.figure.Figure or matplotlib.axes._subplots.AxesSubplot
                The figure and axes of the plot.
        """
        
        if view =='horizon':
            if self.avg_oos_horizon is None:
                self.select_best()
            toplot = self.avg_oos_horizon[score]

        elif view=='cutoff':
            if self.avg_oos_hist is None:
                self.oos_per_cutoff()
            toplot = self.avg_oos_hist[score]

        else:
            raise ValueError('view can take only 2 values: horizon or cutoff')

        if model_subset is not None:
            assert all([m in toplot.columns for m in model_subset]), 'Some models are not in the list of models'
            toplot = toplot[model_subset]
        if ax is None:
            f, ax = plt.subplots(1,1,figsize=figsize)
        if ylabel is None:
            ylabel = score
        if xlabel is None:
            xlabel = view
        
        nseries = len(toplot.columns)
        def expand_color_palette(colors, n_colors_needed):
            base_array = sns.color_palette(colors, n_colors=len(colors))
            color_array = np.array(base_array)
            
            # Interpolating colors
            new_colors = np.vstack([np.interp(np.linspace(0, len(color_array)-1, num=n_colors_needed),
                                            np.arange(len(color_array)), color_array[:, i])
                                    for i in range(color_array.shape[1])]).T
            return new_colors

        # Get the initial colorblind palette
        initial_palette = sns.color_palette("colorblind")

        # Expand the palette if more colors are needed
        if len(initial_palette) < nseries:
            expanded_colors = expand_color_palette("colorblind", n_colors_needed=nseries)
        else:
            expanded_colors = initial_palette

        toplot.plot(ax=ax, style = '-o', color= expanded_colors)
        ax.set(xlabel = xlabel, ylabel=ylabel)
        ax.set_title(title, size="xx-large")
        ax.legend(frameon=False, bbox_to_anchor=(1, 1))
        if ax is None:
            return f, ax
        else:
            return ax

    def plot_prediction(self, 
                        y_pred: pd.Series, 
                        models_preds: Optional[pd.DataFrame] = None, 
                        y_pred_interval: Optional[Tuple] = None, 
                        interval_label: str = 'CI', 
                        aggregation_label: str = 'Model Agg', 
                        xlabel: str = '', 
                        ylabel: str = '', 
                        title: str = 'Prediction', 
                        ax: Optional[plt.Axes] = None, 
                        figsize: Tuple[int, int] = (15, 6)

                       ) -> Union[plt.Figure, plt.Axes]:
        """
        Plot forecast predictions and aggregated prediction.

        Parameters:
        -----------        
            y_pred : pd.Series
                The aggregated prediction.
            models_preds : pd.DataFrame, optional
                The DataFrame containing predictions from different models. Default is None.
            y_pred_interval : tuple, optional
                Prediction interval for the aggregated prediction. Default is None.
            interval_label : str, optional
                Label for the prediction interval. Default is 'prediction interval'.
            aggregation_label : str, optional
                Label for the aggregated prediction. Default is 'Model Agg'.
            xlabel : str, optional
                Label for the x-axis. Default is an empty string.
            ylabel : str, optional
                Label for the y-axis. Default is an empty string.
            title : str, optional
                Title of the plot. Default is 'Prediction'.
            ax : matplotlib.axes._subplots.AxesSubplot, optional
                Matplotlib axes to use for plotting. Default is None.
            figsize : tuple, optional
                Size of the figure in inches (width, height). Default is (15, 6).

        Returns:
        --------        
            matplotlib.figure.Figure or matplotlib.axes._subplots.AxesSubplot
                The figure and axes of the plot.
        """

        y = list(self.LF_d.values())[0].y
        y_train = y.loc[y.index<y_pred.index[0]]
        zoom_y_train = y_train.iloc[-3*len(y_pred):]
    
        if models_preds is not None:
            model_names = models_preds.columns
            models_preds = [models_preds[c] for c in model_names]
            labels = ['y'] + list(model_names) + [aggregation_label]

            plt_res = plot_series(zoom_y_train, *models_preds, y_pred,
                               labels =labels,
                               xlabel =xlabel,
                               ylabel =ylabel,
                               title =title,
                               ax=ax, 
                               figsize =figsize,
                               pred_interval = y_pred_interval,
                               interval_label =interval_label)
        else:
            plt_res = plot_series(zoom_y_train, y_pred,
                               labels =['y', aggregation_label],
                               xlabel =xlabel,
                               ylabel =ylabel,
                               title =title,
                               ax=ax, 
                               figsize =figsize,
                               pred_interval = y_pred_interval,
                              interval_label =interval_label)
        if ax is None:
            f, ax = plt_res
            ax.legend(frameon=False, bbox_to_anchor=(1, 1))
            return f, ax
        else:
            ax = plt_res
            ax.legend(frameon=False, bbox_to_anchor=(1, 1))
            return ax

    def plot_train_test(self, **kwargs: Any):
        """
        Plot the training and test windows for each model.

        Parameters:
        -----------        
            kwargs
                Additional keyword arguments to be passed to the plot function.

        Returns:
        --------        
            fig : plt.Figure
                If ax was None, a new figure is created and returned
                If ax was not None, the same ax is returned with plot added
            ax : plt.Axis             
                Axes containing the plot             
        """
        _lf  = list(self.LF_d.values())[0]
        return _lf.plot.plot_train_test(**kwargs)
    
    def plot_cv_procedure(self,**kwargs: Any):
        """
        Plot the cross-validation procedure for each model.

        Parameters:
        -----------        
            kwargs
                Additional keyword arguments to be passed to the plot function.

        Returns:
        --------
            fig : plt.Figure
                If ax was None, a new figure is created and returned
                If ax was not None, the same ax is returned with plot added
            ax : plt.Axis             
                Axes containing the plot             
        """
        _lf  = list(self.LF_d.values())[0]
        return _lf.plot.plot_cv_procedure(**kwargs)
    
    def __aggregate_pred(self, mode, preds, pred_ints, score=None, ret_underlying=False):

        if score is None:
            score = self.score
 
        nbest = self.nbest
        if mode is None:
            mode = self.mode

        if  (mode !='average') & (self.best_x_overall is None):
            self.select_best(score = score)

        if mode =='best':
            # returns the prediction of the best model
            _best_model = self.best_x_overall[score][0]
            y_pred = preds[_best_model]            
            if _best_model in pred_ints.columns:
                y_pred_int = pred_ints[_best_model]
            else:
                print(f'Prediction intervals for {_best_model} are not available')
                y_pred_int = pd.DataFrame()

        elif mode =='best_horizon':
            # returns the prediction of the best model for each forecast horizon
            best_horizon = self.model_rank_perhorizon[score].loc['Best_1'].to_dict()
            y_pred = pd.Series([preds[v].iloc[k-1] for k, v in best_horizon.items()], 
                   index=preds.index)
            if all([v in pred_ints.columns for v in best_horizon.values()]):
                y_pred_int = pd.concat([pred_ints[v].iloc[k-1].to_frame().T for k, v in best_horizon.items()])
            else:
                print('One or several model that are best in their respective horizon are not available')
                y_pred_int = pd.DataFrame() 

        elif mode == 'average':
            # returns the average prediction of all models
            y_pred = preds.mean(axis=1)
            y_pred_int = pred_ints.unstack().unstack(0).mean(axis=1).unstack().T

        elif mode == 'inverse_score':
            # returns the Average models proportionals to performance score
            _score = self.summary_results.loc[f'Avg {score}'].sort_values()
            _score = _score/_score.sum()

            _weigh_preds =  preds.mul(_score, axis=1).sum(axis=1)
            y_pred = _weigh_preds.astype('float')

            _weigh_preds_ints = 0
            for m, w in _score.to_dict().items():
                if m not in pred_ints.columns:
                    print(f'Prediction intervals for model {m} are not available.') 
                    print('It will be droped in the aggregation of prediction intervals')
                    continue
                else:
                    _weigh_preds_ints = _weigh_preds_ints+ pred_ints[m]*w
            y_pred_int = _weigh_preds_ints.astype('float')

        elif mode == 'nbest_average':
            # return the average prediction of nbest models
            y_pred = preds[self.best_x_overall[score]].mean(axis=1)
            if all([v in pred_ints.columns for v in self.best_x_overall[score]]):
                y_pred_int = pred_ints[self.best_x_overall[score]]\
            .unstack().unstack(0).mean(axis=1).unstack().T
            else:
                print(f'Prediction intervals for {nbest} best models are not available')
                y_pred_int = None

        elif mode == 'nbest_average_horizon':
            # return the average prediction on the nbest models per horizon
            best_horizon = self.model_rank_perhorizon[score].iloc[:nbest].T
            best_horizon = best_horizon.apply(lambda x: x.values, axis=1).to_dict()

            y_pred = pd.Series([preds[v].iloc[k-1].mean() for k, v in best_horizon.items()], 
                               index=preds.index)
            if all([v in pred_ints.columns for v in best_horizon.values()]):
                y_pred_int = pd.concat([pred_ints[v].iloc[k-1].unstack(0).mean(axis=1).to_frame().T for k, v in best_horizon.items()])
                y_pred_int.index = y_pred.index        
            else:
                print(f'Prediction intervals for {nbest} best models are not available')
                y_pred_int = None
                
        else:
            recog_modes = ['best', 'best_horizon', 'average', 'inverse_score', 
                           'nbest_average', 'nbest_average_horizon']
            _error_msg = f'Aggregation mode not recognized. Recognized prediction aggregation are {recog_modes}'
            raise ValueError(_error_msg)

        if ret_underlying:
            return y_pred, y_pred_int, preds, pred_ints
        else:
            return y_pred, y_pred_int




"""Hierarchical Time Series Forecast Reconciliation."""
__description__ = "Time Series Forecast Combination"
__author__ = "Amine Raboun - amineraboun@github.io"

import logging
logging.getLogger().setLevel(logging.ERROR)
import warnings
warnings.simplefilter('ignore')
warnings.filterwarnings("ignore")

import numpy as np
import pandas as pd
from numpy.linalg import pinv
from typing import Dict, Optional

from .forecast import Forecast
from .model_select import ForecastModelSelect

# Plotting packages & configuration
from forecast_combine.utils.plotting import plot_series
import matplotlib.pyplot as plt

class ForecastReconciler:
    def __init__(self, 
                 forecasters_d: Dict[str, any], 
                 S: pd.DataFrame, 
                 method: str = 'mint',
                 ):
        """
        Initializes the ForecastReconciler class.
        
        Parameters:
        -----------
        - forecasters_d (Dict[str, any]): Dictionary of forecaster objects for each series.
        - S (pd.DataFrame): Aggregation matrix indicating how lower levels aggregate into the total.
        - method (str): Method of reconciliation ('ols', 'wls', 'td', 'mint', 'bu').
        """
        assert S.shape[0] == len(forecasters_d), "Number of forecasters must match number of rows in S."
        assert S.shape[1] == S.shape[0]-1, "Number of columns in S must match number of the child forecasters."
        assert all([s in forecasters_d.keys() for s in S.index]), "Index values of S must match keys in forecasters_d."
        self.methods_l = ['ols', 'wls', 'td', 'mint', 'bu']
        assert method in self.methods_l, "Invalid reconciliation method."
        self.forecasters_d = forecasters_d
        self.Total = S.index[0]
        self.forecasters_order = S.index
        self.S = S.values
        self.method = method
        self.historical_values_d = None
        self.historical_errors_d = None
        self.weights = None
   
    def fit(self,
            refit_models: bool = False,
            reconciliation_method: Optional[str] = None
            ) -> None:
        """
        Prepare the reconciliation weights or parameters based on historical data.
        Optionally refit models.
        """
        if refit_models:
            for forecaster in self.forecasters_d.values():
                forecaster.fit()
        if reconciliation_method is None:
            reconciliation_method = self.method

        if reconciliation_method in ['wls', 'mint']:
            print(f"{reconciliation_method} reconciliation method requires the forecasters prediction errors variance-covariance matrix ...")
            self.historical_errors_d = self.__compute_hist_errors()
            historical_errors = pd.concat(self.historical_errors_d).reset_index(0).rename(columns={'level_0':'nodes'})

            if reconciliation_method == 'wls':
                _wght = {}
                for horizon, group in historical_errors.groupby('horizon'):
                    _var = group.pivot(index='cutoff', columns='nodes', values='error').var()
                    _wght[horizon] = np.diag(1 / _var.loc[self.forecasters_order])
                self.weights = _wght
            elif reconciliation_method == 'mint':
                _wght = {}
                for horizon, group in historical_errors.groupby('horizon'):
                    ordered_group = group.pivot(index='cutoff', columns='nodes', values='error')[self.forecasters_order]
                    _wght[horizon] = pinv(ordered_group.cov().values)
                self.weights = _wght
        elif reconciliation_method == 'td':
            self.historical_values_d = self.__compute_hist_values()
            historical_values_df = pd.concat(self.historical_values_d, axis=1)[self.forecasters_order]
            historical_values_prop = (historical_values_df.iloc[:, 1:]*self.S[0]).div(historical_values_df[self.Total], axis=0)
            historical_values_prop_mean = historical_values_prop.mean().to_dict()
            historical_values_prop_mean[self.Total] = 1
            self.weights = historical_values_prop_mean
        else:
            pass  # No weights needed for 'bu' or 'ols'
   
    def predict(self, 
                X: Dict[str, pd.DataFrame] = None,
                coverage: float = 0.9,
                reconciliation_method: Optional[str] = None,
                verbose: bool = False
                ) -> pd.DataFrame:
        """
        Apply the reconciliation method to the forecasts.
        
        Parameters:
        -----------
        - X (Optional[pd.DataFrame]): DataFrame containing independent variables for prediction, if needed.
        - coverage (float): Confidence interval for the forecast quantiles.
        - reconciliation_method (Optional[str]): Override the default method for this prediction.
        
        Returns:
        --------
        - pd.DataFrame: Reconciled forecasts.
        """
        assert reconciliation_method in self.methods_l, "Invalid reconciliation method."
        if verbose:
            print("Computing the Hiearchical models forecasts ...")
        forecast_data = {}
        forecast_intervals ={}
        for name, forecaster in self.forecasters_d.items():
            if X is None:
                forecster_X = None
            else:
                forecster_X = X[name] if name in X.keys() else None
            if verbose:
                print(f"\tComputing forecasts for {name} ...")
            forecast_data[name], forecast_intervals[name] = forecaster.predict(X=forecster_X, coverage=coverage)

        if verbose:
            print("\nReconciling forecasts ...")
        return self.reconcile_preds(forecast_data, forecast_intervals, reconciliation_method)
    
    def update(self,
               new_y:Dict[str, pd.Series],
               new_X: Dict[str, pd.DataFrame] = None,
               reconciliation_method: Optional[str] = None,
               coverage: float = 0.9, 
               refit: bool = False, 
               reevaluate:bool = False, 
               verbose: bool = False
              ) -> pd.DataFrame: 
        """
        Update the forecast with new data.

        Parameters:
        -----------
        - newdata (pd.DataFrame): New data to update the forecast.
        - reconciliation_method (Optional[str]): Override the default method for this prediction.
        - refit (bool): If True, refit the models before updating the forecast.

        Returns:
        --------
        - pd.DataFrame: Reconciled forecasts.
        """        
        assert reconciliation_method in self.methods_l, "Invalid reconciliation method."   
        if verbose:
            print("Computing the Hiearchical models forecasts ...")
        forecast_data = {}
        forecast_intervals = {}
        for name, forecaster in self.forecasters_d.items():
            forecaster_y = new_y[name] 
            if new_X is None:
                forecster_X = None
            else:
                forecster_X = new_X[name] if name in new_X.keys() else None
            if verbose:
                print(f"\tComputing forecasts for {name} ...")
            forecast_data[name], forecast_intervals[name] = forecaster.update(new_y=forecaster_y, 
                                                                              new_X=forecster_X,
                                                                              coverage=coverage,
                                                                              refit = refit,
                                                                              reevaluate = reevaluate)
            
        if verbose:
            print("\nReconciling forecasts ...")
        return self.reconcile_preds(forecast_data, forecast_intervals, reconciliation_method)

    def reconcile_preds(self, 
                forecast_data: Dict[str, pd.DataFrame],
                forecast_intervals: Dict[str, pd.DataFrame],
                reconciliation_method: Optional[str] = None
                ) -> pd.DataFrame:
        """
        Apply the reconciliation method to the forecasts.
        
        Parameters:
        -----------
        - forecast_data (Dict[str, pd.DataFrame]): Dictionary of forecasts for each series.
        
        Returns:
        --------
        - pd.DataFrame: Reconciled forecasts.
        """
        if reconciliation_method is None:
            reconciliation_method = self.method
        
        if (reconciliation_method in ['wls', 'mint', 'td']):
            self.fit(reconciliation_method=reconciliation_method)
        
        y_preds = np.vstack([forecast_data[name] for name in self.forecasters_order])
        y_preds_lower = np.vstack([forecast_intervals[name].xs('lower', level =2, axis=1).squeeze() for name in self.forecasters_order])
        y_preds_upper = np.vstack([forecast_intervals[name].xs('upper', level =2, axis=1).squeeze() for name in self.forecasters_order])
        
        # Reconcile forecasts and intervals using the same weights and matrices
        reconciled_forecasts = self.__apply_reconciliation(y_preds, reconciliation_method)
        reconciled_lower_bounds = self.__apply_reconciliation(y_preds_lower, reconciliation_method)
        reconciled_upper_bounds = self.__apply_reconciliation(y_preds_upper, reconciliation_method)

        # Return the reconciled forecasts and intervals
        reconciled_preds ={}
        reconciled_intervals = {}
        dates = forecast_data[self.Total].index
        for i, name in enumerate(self.forecasters_order):
            reconciled_preds[name] = pd.Series(reconciled_forecasts[i], index = dates).rename(name)
            _interval = pd.DataFrame(np.vstack([reconciled_lower_bounds[i], reconciled_upper_bounds[i]]).T, 
                                                      index = dates, columns = ['lower', 'upper'])
            _interval.columns = forecast_intervals[name].columns
            reconciled_intervals[name] = _interval        
        return reconciled_preds, reconciled_intervals
    
    def plot_predict(self, 
                     reconciled_preds,
                     reconciled_intervals,
                     interval_label: str = 'CI', 
                    title: str = 'Reconciled Predictions',
                    ):
        self.historical_values_d = self.__compute_hist_values()
        nplots = len(reconciled_preds)
        ncols = 2
        nrows = nplots // ncols + nplots % ncols
        fig, axes = plt.subplots(nrows, ncols, figsize=(20, 6*nrows))
        axes = axes.flatten()
        for i, name in enumerate(reconciled_preds.keys()):
            y_i = self.historical_values_d[name]
            ypred_i = reconciled_preds[name]
            y_i = y_i.loc[y_i.index<ypred_i.index[0]]
            zoom_y_train = y_i.iloc[-3*len(ypred_i):]
            plot_series(zoom_y_train,
                        ypred_i,
                        labels =['y', 'y_pred'],
                        pred_interval= reconciled_intervals[name],
                        interval_label =interval_label,
                        title =name,
                        ax=axes[i], 
                        )
        fig.tight_layout()
        fig.suptitle(title, size="xx-large")
        fig.subplots_adjust(top=0.9)        
        return fig, axes
    
    def __compute_hist_errors(self, verbose=False):
        if self.historical_errors_d is None:                
            if verbose:
                print("\nComputing the prediction errors for all forecasters ...")
            _d = {}
            for k, forecaster in self.forecasters_d.items():
                if verbose:
                    print(f"\tComputing prediction errors for {k} ...")
                _d[k] = forecaster.get_pred_errors()
            self.historical_errors_d = _d
        return self.historical_errors_d
    def __compute_hist_values(self, verbose=False):    
        if self.historical_values_d is None:
            if verbose:
                print("Computing the historical values for all forecasters ...")
            _d = {}
            for k, forecaster in self.forecasters_d.items():
                if isinstance(forecaster, Forecast):
                    _d[k] = forecaster.y
                elif isinstance(forecaster, ForecastModelSelect):
                    _d[k] = list(forecaster.LF_d.values())[0].y
            self.historical_values_d = _d
        return self.historical_values_d
    
    def __apply_reconciliation(self, y_preds: np.array, reconciliation_method: str) -> pd.DataFrame:
        """
        Apply the reconciliation method to the forecasts.
        """
        if reconciliation_method == 'td':
            total_forecast = y_preds[0]
            reconciled_forecasts_d = {name: total_forecast.T * self.weights[name] for name in self.forecasters_order}
            reconciled_forecasts = reconciled_forecasts_d.values()
            reconciled_forecasts = np.array([total_forecast.T * self.weights[name] for name in self.forecasters_order])  
        elif reconciliation_method == 'bu':
            reconciled_forecasts = self.S @ y_preds[1:, :]
        
        elif reconciliation_method == 'ols':
            M = self.S @ pinv(self.S.T @ self.S) @ self.S.T
            reconciled_forecasts = M  @ y_preds
        
        elif reconciliation_method in ['wls', 'mint']:
            reconciled_forecasts = pd.DataFrame()
            for horizon in self.weights.keys():
                M = self.S @ pinv(self.S.T @ self.weights[horizon] @ self.S) @ self.S.T @ self.weights[horizon]
                reconciled_forecasts = M @ y_preds

        return reconciled_forecasts